/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <gpdef.h>
#include <gpstdlib.h>
#include <gpgraphic.h>
#include <gpstdio.h>
#include <gpfont.h>

#include "fs.h"
#include "types.h"

void fs_init(void) {
	GpFatInit();
	GpRelativePathSet("gp:\\gpmm\\dragon");
}

FS_FILE fs_open(char *filename, int flags) {
	F_HANDLE fd;
	if (flags & FS_WRITE) {
		if (GpFileCreate(filename, ALWAYS_CREATE, &fd) != SM_OK)
			return -1;
	} else {
		if (GpFileOpen(filename, OPEN_R, &fd) != SM_OK)
			return -1;
	}
	return fd;
}

ssize_t fs_read(FS_FILE fd, void *buffer, size_t size) {
	unsigned long count;
	int rbytes = 0;
	ERR_CODE err;
	/* Read in 512 byte chunks, or the GP32 crashes! */
	do {
		err = GpFileRead(fd, buffer, size>512?512:size, &count);
		if (err != SM_OK && err != ERR_EOF)
			return -1;
		size -= count;
		buffer += count;
		rbytes += count;
	} while (err == SM_OK && size > 0);
	return rbytes;
}

ssize_t fs_write(FS_FILE fd, void *buffer, size_t size) {
	ERR_CODE err;
	/* Haven't done any file writing yet, so don't know if similar
	 * contortions are required */
	err = GpFileWrite(fd, buffer, (unsigned long)size);
	if (err != SM_OK)
		return -1;
	return size;
}

void fs_close(FS_FILE fd) {
	GpFileClose(fd);
}

FS_DIR *fs_scandir(char *path, char *ext) {
	FS_DIR *dir = (FS_DIR *)malloc(sizeof(FS_DIR));
	GPDIRENTRY entry;
	GPFILEATTR entry_attr;
	unsigned long p_num, dummy, i, j;

	if (dir == NULL)
		return NULL;
	GpRelativePathSet(path);
	GpDirEnumNum(".", &p_num);
	dir->dirents = (FS_DIRENT *)malloc(p_num * sizeof(FS_DIRENT));
	if (dir->dirents == NULL) {
		free(dir);
		return NULL;
	}
	for (i = 0, j = 0; i < p_num; i++) {
		GpDirEnumList(".", i, 1, &entry, &dummy);
		GpFileAttr(entry.name, &entry_attr);
		memcpy(dir->dirents[j].name, entry.name, 16);
		dir->dirents[j].attr = 0;
		if (entry_attr.attr & 0x10) {
			dir->dirents[j++].attr |= FS_ATTR_DIRECTORY;
		} else {
			if (!strcmp(entry.name+strlen(entry.name)-3,ext))
				j++;
		}
	}
	dir->num_files = j;
	return dir;
}

void fs_freedir(FS_DIR *dir) {
	free(dir->dirents);
	free(dir);
}

ssize_t fs_load_file(char *filename, void *buf, size_t size) {
	int count;
	FS_FILE fd;
	if ((fd = fs_open(filename, FS_READ)) == -1)
		return -1;
	count = fs_read(fd, buf, size);
	fs_close(fd);
	return count;
}

int fs_write_byte(FS_FILE fd, uint8_t octet) {
	return fs_write(fd, &octet, 1);
}

int fs_write_word16(FS_FILE fd, uint16_t word16) {
	if (fs_write_byte(fd, word16 >> 8) == -1)
		return -1;
	return fs_write_byte(fd, word16 & 0xff);
}

int fs_write_word32(FS_FILE fd, uint32_t word32) {
	if (fs_write_word16(fd, word32 >> 16) == -1)
		return -1;
	return fs_write_word16(fd, word32 & 0xffff);
}

int fs_read_byte(FS_FILE fd, uint8_t *dest) {
	return fs_read(fd, dest, 1);
}

int fs_read_word16(FS_FILE fd, uint16_t *dest) {
	uint8_t octet;
	int ret;
	if (fs_read(fd, &octet, 1) == -1)
		return -1;
	*dest = octet << 8;
	if ((ret = fs_read(fd, &octet, 1)) == -1)
		return -1;
	*dest |= octet;
	return ret;
}

int fs_read_word32(FS_FILE fd, uint32_t *dest) {
	uint16_t word16;
	int ret;
	if (fs_read_word16(fd, &word16) == -1)
		return -1;
	*dest = word16 << 16;
	if ((ret = fs_read_word16(fd, &word16)) == -1)
		return -1;
	*dest |= word16;
	return ret;
}
