/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_SDL_AUDIO

#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <SDL.h>

#include "sound.h"
#include "pia.h"
#include "xroar.h"
#include "logging.h"
#include "types.h"

typedef Uint8 Sample;  /* 8-bit mono (SDL type) */
typedef uint_fast8_t Sample_f;  /* Fastest for manipulating above */

#define SAMPLE_RATE 22050
#define CHANNELS 1
#define FORMAT AFMT_U8
#define FRAGMENTS 2
#define FRAME_SIZE 512
#define SAMPLE_CYCLES ((uint32_t)(OSCILLATOR_RATE / SAMPLE_RATE))
#define FRAME_CYCLES (SAMPLE_CYCLES * FRAME_SIZE)

static SDL_AudioSpec audiospec;
static Cycle frame_cycle_base;
static Sample *buffer;
static Sample *wrptr;
static Sample_f lastsample;
static volatile uint_fast8_t haltflag;

static void callback(void *userdata, Uint8 *stream, int len);

int sound_init(void) {
	LOG_DEBUG(2,"Initialising SDL audio driver\n");
	SDL_InitSubSystem(SDL_INIT_AUDIO);
	audiospec.freq = SAMPLE_RATE;
	audiospec.format = AUDIO_U8;
	audiospec.samples = FRAME_SIZE;
	audiospec.channels = CHANNELS;
	audiospec.callback = callback;
	audiospec.userdata = NULL;
	if (SDL_OpenAudio(&audiospec, NULL) < 0) {
		LOG_ERROR("Couldn't open audio: %s\n", SDL_GetError());
		return -1;
	}
	SDL_PauseAudio(0);
	buffer = (Sample *)malloc(FRAME_SIZE * sizeof(Sample));
	return 0;
}

void sound_shutdown(void) {
	LOG_DEBUG(2,"Shutting down SDL audio driver\n");
	SDL_CloseAudio();
	SDL_QuitSubSystem(SDL_INIT_AUDIO);
	free(buffer);
}

void sound_reset(void) {
	memset(buffer, 0x00, FRAME_SIZE * sizeof(Sample));
	wrptr = buffer;
	frame_cycle_base = current_cycle;
	next_sound_update = frame_cycle_base + FRAME_CYCLES;
	lastsample = 0;
}

void sound_update(void) {
	Cycle elapsed_cycles = current_cycle - frame_cycle_base;
	Sample_f fill_with;
	Sample *fill_to;
	if (elapsed_cycles >= FRAME_CYCLES) {
		fill_to = buffer + FRAME_SIZE;
	} else {
		fill_to = buffer + (elapsed_cycles/(Cycle)SAMPLE_CYCLES);
	}
	if (!(PIA_1B.control_register & 0x08)) {
		/* Single-bit sound */
		//fill_with = (PIA_1B.port_output & 0x02) << 6;
		fill_with = lastsample;
	} else  {
		if (PIA_0B.control_register & 0x08 || PIA_0A.control_register & 0x08) {
			/* Sound disabled */
			fill_with = lastsample;
		} else {
			/* DAC output */
			fill_with = PIA_1A.port_output & 0xfc;
		}
	}
	while (haltflag);
	while (wrptr < fill_to)
		*(wrptr++) = lastsample;
	if ((fill_to - buffer) >= FRAME_SIZE) {
		frame_cycle_base += FRAME_CYCLES;
		next_sound_update = frame_cycle_base + FRAME_CYCLES;
		wrptr = buffer;
		SDL_LockAudio();
		haltflag = 1;
		SDL_UnlockAudio();
	}
	lastsample = fill_with;
	return;
}

static void callback(void *userdata, Uint8 *stream, int len) {
	//LOG_DEBUG(0,"FRAME_SIZE = %d, len = %d\n", FRAME_SIZE, len);
	if (len == FRAME_SIZE) {
		memcpy(stream, buffer, len);
		//SDL_MixAudio(stream, buffer, len, SDL_MIX_MAXVOLUME);
	}
	haltflag = 0;
}

#endif  /* HAVE_SDL_AUDIO */
