/*  XRoar - a Dragon/Tandy Coco emulator
 *  Copyright (C) 2003-2004  Ciaran Anscomb
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

/* This file contains VDG routines rewritten for GP32.  They process 4
 * scanlines at once and write to the video memory in aligned 32-bit words
 * thus reducing significantly the amount of writes to uncached RAM and
 * speeding things up a lot. */

#include "video.h"
#include "m6809.h"
#include "sam.h"
#include "vdg.h"
#include "pia.h"
#include "logging.h"
#include "types.h"

#include "vdg_bitmaps.c"
#include "vdg_bitmaps_gp32.c"

static void render_sg4(void);
/* TODO: static void render_sg6(void); */
static void render_cg1(void);
static void render_rg1(void);
static void render_cg2(void);
static void render_rg6(void);
void (*vdg_render_scanline)(void) = render_sg4;

static uint_fast8_t subline;

static uint32_t *pixel;
static uint32_t darkgreen, black;
static Pixel bg_colour;
static Pixel fg_colour;
static Pixel vdg_colour[8];
static Pixel *cg_colours;
static uint32_t border_colour;
static uint8_t *rendered_alpha;

void vdg_init(void) {
	vdg_colour[0] = MAPCOLOUR(0x00, 0xff, 0x00);
	vdg_colour[1] = MAPCOLOUR(0xff, 0xff, 0x00);
	vdg_colour[2] = MAPCOLOUR(0x00, 0x00, 0xff);
	vdg_colour[3] = MAPCOLOUR(0xff, 0x00, 0x00);
	vdg_colour[4] = MAPCOLOUR(0xff, 0xe0, 0xe0);
	vdg_colour[5] = MAPCOLOUR(0x00, 0xff, 0xff);
	vdg_colour[6] = MAPCOLOUR(0xff, 0x00, 0xff);
	vdg_colour[7] = MAPCOLOUR(0xff, 0xa5, 0x00);
	black = MAPCOLOUR(0x00, 0x00, 0x00);
	black = (black<<24) | (black<<16) | (black<<8) | black;
	darkgreen = MAPCOLOUR(0x00, 0x20, 0x00);
	darkgreen = (darkgreen<<24) | (darkgreen<<16) | (darkgreen<<8) | darkgreen;
}

void vdg_reset(void) {
	vdg_set_mode();
	pixel = (uint32_t *)screen.ptbuffer + 58;
	subline = 0;
}

void vdg_vsync(void) {
	VIDEO_UPDATE;
	sam_vdg_fsync();
	pixel = (uint32_t *)screen.ptbuffer + 58;
	subline = 0;
}

static inline void vram_ptrs_16(uint8_t **ptrs) {
	int i;
	for (i = 4; i; i--) {
		*(ptrs++) = (uint8_t *)sam_vram_ptr(sam_vdg_address);
		sam_vdg_xstep(16);
		sam_vdg_hsync(16,6,0);
	}
}

static inline void vram_ptrs_32(uint8_t **ptrs) {
	int i, j;
	for (j = 4; j; j--) {
		for (i = 2; i; i--) {
			*(ptrs++) = (uint8_t *)sam_vram_ptr(sam_vdg_address);
			sam_vdg_xstep(16);
		}
		sam_vdg_hsync(32,10,16);
	}
}

static void render_sg4(void) {
	uint8_t *vram0, *vram1, *vram2, *vram3;
	uint32_t in4, in5, in6, in7, out, i, j, k;
	uint32_t *dest;
	uint8_t *charset;
	uint8_t *vram_ptrs[8];
	dest = pixel--;
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	charset = rendered_alpha + subline * 8192;
	vram_ptrs_32(vram_ptrs);
	for (k = 0; k < 2; k++) {
		vram0 = vram_ptrs[k]; vram1 = vram_ptrs[2+k];
		vram2 = vram_ptrs[4+k]; vram3 = vram_ptrs[6+k];
		for (j = 16; j; j--) {
			in4 = (*(vram0++) << 2);
			in5 = (*(vram1++) << 2) + 1;
			in6 = (*(vram2++) << 2) + 2;
			in7 = (*(vram3++) << 2) + 3;
			for (i = 8; i; i--) {
				out = *(charset + (in4<<3)) << 24;
				out |= *(charset + (in5<<3)) << 16;
				out |= *(charset + (in6<<3)) << 8;
				out |= *(charset + (in7<<3));
				*dest = out;
				dest += 60;
				charset++;
			}
			charset -= 8;
		}
	}
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	subline++;
	if (subline > 2) subline = 0;
}

static void render_cg1(void) {
	uint8_t *vram0, *vram1, *vram2, *vram3;
	uint32_t in4, in5, in6, in7, out, i, j;
	uint32_t *dest;
	uint8_t *vram_ptrs[4];
	dest = pixel--;
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	vram_ptrs_16(vram_ptrs);
	vram0 = vram_ptrs[0]; vram1 = vram_ptrs[1];
	vram2 = vram_ptrs[2]; vram3 = vram_ptrs[3];
	for (j = 16; j; j--) {
		in4 = *(vram0++);
		in5 = *(vram1++);
		in6 = *(vram2++);
		in7 = *(vram3++);
		out = *(cg_colours + ((in4&0xc0)>>6)) << 24;
		out |= *(cg_colours + ((in5&0xc0)>>6)) << 16;
		out |= *(cg_colours + ((in6&0xc0)>>6)) << 8;
		out |= *(cg_colours + ((in7&0xc0)>>6));
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = *(cg_colours + ((in4&0x30)>>4)) << 24;
		out |= *(cg_colours + ((in5&0x30)>>4)) << 16;
		out |= *(cg_colours + ((in6&0x30)>>4)) << 8;
		out |= *(cg_colours + ((in7&0x30)>>4));
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = *(cg_colours + ((in4&0x0c)>>2)) << 24;
		out |= *(cg_colours + ((in5&0x0c)>>2)) << 16;
		out |= *(cg_colours + ((in6&0x0c)>>2)) << 8;
		out |= *(cg_colours + ((in7&0x0c)>>2));
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = *(cg_colours + (in4&0x03)) << 24;
		out |= *(cg_colours + (in5&0x03)) << 16;
		out |= *(cg_colours + (in6&0x03)) << 8;
		out |= *(cg_colours + (in7&0x03));
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		*dest = out; dest += 60;
	}
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	subline++;
	if (subline > 2) subline = 0;
}

static void render_rg1(void) {
	uint8_t *vram0, *vram1, *vram2, *vram3;
	uint32_t in4, in5, in6, in7, out, i, j;
	uint32_t *dest;
	uint8_t *vram_ptrs[4];
	dest = pixel--;
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	vram_ptrs_16(vram_ptrs);
	vram0 = vram_ptrs[0]; vram1 = vram_ptrs[1];
	vram2 = vram_ptrs[2]; vram3 = vram_ptrs[3];
	for (j = 16; j; j--) {
		in4 = *(vram0++);
		in5 = *(vram1++);
		in6 = *(vram2++);
		in7 = *(vram3++);
		out = ((in4 & 0x80)?fg_colour:bg_colour) << 24;
		out |= ((in5 & 0x80)?fg_colour:bg_colour) << 16;
		out |= ((in6 & 0x80)?fg_colour:bg_colour) << 8;
		out |= ((in7 & 0x80)?fg_colour:bg_colour);
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = ((in4 & 0x40)?fg_colour:bg_colour) << 24;
		out |= ((in5 & 0x40)?fg_colour:bg_colour) << 16;
		out |= ((in6 & 0x40)?fg_colour:bg_colour) << 8;
		out |= ((in7 & 0x40)?fg_colour:bg_colour);
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = ((in4 & 0x20)?fg_colour:bg_colour) << 24;
		out |= ((in5 & 0x20)?fg_colour:bg_colour) << 16;
		out |= ((in6 & 0x20)?fg_colour:bg_colour) << 8;
		out |= ((in7 & 0x20)?fg_colour:bg_colour);
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = ((in4 & 0x10)?fg_colour:bg_colour) << 24;
		out |= ((in5 & 0x10)?fg_colour:bg_colour) << 16;
		out |= ((in6 & 0x10)?fg_colour:bg_colour) << 8;
		out |= ((in7 & 0x10)?fg_colour:bg_colour);
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = ((in4 & 0x08)?fg_colour:bg_colour) << 24;
		out |= ((in5 & 0x08)?fg_colour:bg_colour) << 16;
		out |= ((in6 & 0x08)?fg_colour:bg_colour) << 8;
		out |= ((in7 & 0x08)?fg_colour:bg_colour);
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = ((in4 & 0x04)?fg_colour:bg_colour) << 24;
		out |= ((in5 & 0x04)?fg_colour:bg_colour) << 16;
		out |= ((in6 & 0x04)?fg_colour:bg_colour) << 8;
		out |= ((in7 & 0x04)?fg_colour:bg_colour);
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = ((in4 & 0x02)?fg_colour:bg_colour) << 24;
		out |= ((in5 & 0x02)?fg_colour:bg_colour) << 16;
		out |= ((in6 & 0x02)?fg_colour:bg_colour) << 8;
		out |= ((in7 & 0x02)?fg_colour:bg_colour);
		*dest = out; dest += 60;
		*dest = out; dest += 60;
		out = ((in4 & 0x01)?fg_colour:bg_colour) << 24;
		out |= ((in5 & 0x01)?fg_colour:bg_colour) << 16;
		out |= ((in6 & 0x01)?fg_colour:bg_colour) << 8;
		out |= ((in7 & 0x01)?fg_colour:bg_colour);
		*dest = out; dest += 60;
		*dest = out; dest += 60;
	}
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	subline++;
	if (subline > 2) subline = 0;
}

static void render_cg2(void) {
	uint8_t *vram0, *vram1, *vram2, *vram3;
	uint32_t in4, in5, in6, in7, out, i, j, k;
	uint32_t *dest;
	uint8_t *vram_ptrs[8];
	dest = pixel--;
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	vram_ptrs_32(vram_ptrs);
	for (k = 0; k < 2; k++) {
		vram0 = vram_ptrs[k]; vram1 = vram_ptrs[2+k];
		vram2 = vram_ptrs[4+k]; vram3 = vram_ptrs[6+k];
		for (j = 16; j; j--) {
			in4 = *(vram0++);
			in5 = *(vram1++);
			in6 = *(vram2++);
			in7 = *(vram3++);
			out = *(cg_colours + ((in4&0xc0)>>6)) << 24;
			out |= *(cg_colours + ((in5&0xc0)>>6)) << 16;
			out |= *(cg_colours + ((in6&0xc0)>>6)) << 8;
			out |= *(cg_colours + ((in7&0xc0)>>6));
			*dest = out;
			dest += 60;
			*dest = out;
			dest += 60;
			out = *(cg_colours + ((in4&0x30)>>4)) << 24;
			out |= *(cg_colours + ((in5&0x30)>>4)) << 16;
			out |= *(cg_colours + ((in6&0x30)>>4)) << 8;
			out |= *(cg_colours + ((in7&0x30)>>4));
			*dest = out;
			dest += 60;
			*dest = out;
			dest += 60;
			out = *(cg_colours + ((in4&0x0c)>>2)) << 24;
			out |= *(cg_colours + ((in5&0x0c)>>2)) << 16;
			out |= *(cg_colours + ((in6&0x0c)>>2)) << 8;
			out |= *(cg_colours + ((in7&0x0c)>>2));
			*dest = out;
			dest += 60;
			*dest = out;
			dest += 60;
			out = *(cg_colours + (in4&0x03)) << 24;
			out |= *(cg_colours + (in5&0x03)) << 16;
			out |= *(cg_colours + (in6&0x03)) << 8;
			out |= *(cg_colours + (in7&0x03));
			*dest = out;
			dest += 60;
			*dest = out;
			dest += 60;
		}
	}
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	subline++;
	if (subline > 2) subline = 0;
}

static void render_rg6(void) {
	uint8_t *vram0, *vram1, *vram2, *vram3;
	uint32_t in4, in5, in6, in7, out, i, j, k;
	uint32_t *dest;
	uint8_t *vram_ptrs[8];
	dest = pixel--;
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	vram_ptrs_32(vram_ptrs);
	for (k = 0; k < 2; k++) {
		vram0 = vram_ptrs[k]; vram1 = vram_ptrs[2+k];
		vram2 = vram_ptrs[4+k]; vram3 = vram_ptrs[6+k];
		for (j = 16; j; j--) {
			in4 = *(vram0++);
			in5 = *(vram1++);
			in6 = *(vram2++);
			in7 = *(vram3++);
			out = ((in4 & 0x80)?fg_colour:bg_colour) << 24;
			out |= ((in5 & 0x80)?fg_colour:bg_colour) << 16;
			out |= ((in6 & 0x80)?fg_colour:bg_colour) << 8;
			out |= ((in7 & 0x80)?fg_colour:bg_colour);
			*dest = out; dest += 60;
			out = ((in4 & 0x40)?fg_colour:bg_colour) << 24;
			out |= ((in5 & 0x40)?fg_colour:bg_colour) << 16;
			out |= ((in6 & 0x40)?fg_colour:bg_colour) << 8;
			out |= ((in7 & 0x40)?fg_colour:bg_colour);
			*dest = out; dest += 60;
			out = ((in4 & 0x20)?fg_colour:bg_colour) << 24;
			out |= ((in5 & 0x20)?fg_colour:bg_colour) << 16;
			out |= ((in6 & 0x20)?fg_colour:bg_colour) << 8;
			out |= ((in7 & 0x20)?fg_colour:bg_colour);
			*dest = out; dest += 60;
			out = ((in4 & 0x10)?fg_colour:bg_colour) << 24;
			out |= ((in5 & 0x10)?fg_colour:bg_colour) << 16;
			out |= ((in6 & 0x10)?fg_colour:bg_colour) << 8;
			out |= ((in7 & 0x10)?fg_colour:bg_colour);
			*dest = out; dest += 60;
			out = ((in4 & 0x08)?fg_colour:bg_colour) << 24;
			out |= ((in5 & 0x08)?fg_colour:bg_colour) << 16;
			out |= ((in6 & 0x08)?fg_colour:bg_colour) << 8;
			out |= ((in7 & 0x08)?fg_colour:bg_colour);
			*dest = out; dest += 60;
			out = ((in4 & 0x04)?fg_colour:bg_colour) << 24;
			out |= ((in5 & 0x04)?fg_colour:bg_colour) << 16;
			out |= ((in6 & 0x04)?fg_colour:bg_colour) << 8;
			out |= ((in7 & 0x04)?fg_colour:bg_colour);
			*dest = out; dest += 60;
			out = ((in4 & 0x02)?fg_colour:bg_colour) << 24;
			out |= ((in5 & 0x02)?fg_colour:bg_colour) << 16;
			out |= ((in6 & 0x02)?fg_colour:bg_colour) << 8;
			out |= ((in7 & 0x02)?fg_colour:bg_colour);
			*dest = out; dest += 60;
			out = ((in4 & 0x01)?fg_colour:bg_colour) << 24;
			out |= ((in5 & 0x01)?fg_colour:bg_colour) << 16;
			out |= ((in6 & 0x01)?fg_colour:bg_colour) << 8;
			out |= ((in7 & 0x01)?fg_colour:bg_colour);
			*dest = out; dest += 60;
		}
	}
	for (i = 32; i; i--) {
		*dest = border_colour;
		dest += 60;
	}
	subline++;
	if (subline > 2) subline = 0;
}

void vdg_set_mode(void) {
	uint_fast8_t mode = PIA_1B.port_output;
	if (mode & 0x80) {
		/* Graphics modes */
		cg_colours = &vdg_colour[(mode & 0x08) >> 1];
		bg_colour = black;
		fg_colour = cg_colours[0];
		border_colour = (fg_colour<<24)|(fg_colour<<16)|(fg_colour<<8)|fg_colour;
		switch ((mode & 0x70) >> 4) {
			case 0: vdg_render_scanline = render_cg1; break;
			case 1:
			case 3:
			case 5: vdg_render_scanline = render_rg1; break;
			case 2:
			case 4:
			case 6: vdg_render_scanline = render_cg2; break;
			default:
			case 7: vdg_render_scanline = render_rg6; break;
		}
	} else {
		border_colour = black;
		vdg_render_scanline = render_sg4;
		rendered_alpha = (uint8_t *)vdg_alpha_gp32[(mode & 0x08) >> 3];
	}
}
